package ASN1tokenizer;

use strict;
use BooleanTF;
use ASN1Token;


###############################
# How to use this module
###############################
#
# Your driver script must first call "initFromStdin()".  This function reads the entire
# ASN1 file from STDIN and converts it to an array of tokens.
#
# After this, your driver can call getToken().  This returns an object of type ASN1Token.
# This type has 3 data elements:
#   1) "line"   - line number from the input line where this token was found
#   2) "type"   - type of token: TOKEN or BLOCK_COMMENT or EOL_COMMENT.  See below.
#   3) "string" - the text of a token or comment.  See below.
#
# Definitions of each token type:
#   TOKEN: any of { } ( ) , ::= ..
#          or a word consisting of any non-whitespace characters and none of above characters or strings
#   BLOCK_COMMENT:
#          Multiple adjacent lines of whole-line comments (and also any trailing empty lines) are
#          saved into one token of this type.
#   EOL_COMMENT:
#          Indicates that the comment follows some tokens on the same line.
#
# Constants for these token types are found in the ASN1Token module:
#     ASN1Token::TOKEN         (equal to 0)
#     ASN1Token::BLOCK_COMMENT (equal to 1)
#     ASN1Token::EOL_COMMENT   (equal to 2)
# If Perl had enums, this would be one.
#
#
#





#------------------------------------------------------------------------------
# "PUBLIC" PROTOTYPES
# Driver programs should use these functions and nothing else from this file.
# They are not exported, so call them like "ASN1tokenizer::function()".
#------------------------------------------------------------------------------
sub initFromStdin();
sub getToken();
sub peek(;$);  #returns the nth upcoming token (default n=1)



#==============================================
# Everything below here is "private".
# Users of this module should only use the interface above.
#==============================================

#----------------------
# vars
#----------------------
my @_tokens;  #all the so-far extracted tokens
my $_inputline = 0; #counter

#----------------------
# private prototypes
#----------------------
sub extractTokensFromLine($);


#----------------------
# function definitions
#----------------------

sub initFromStdin()
{
    print "========================================================\n";
    print " TOKENIZING from the input file\n";
    print " (Skipping Word header prose and section header lines)\n";
    print "========================================================\n";

    # When this var is true, ignore input
    my $inTokens = FALSE;
    
    while( my $line=<> )
    {
        $_inputline++;

        # Fix strange character translations from copy/pasting the spec from Word
        $line =~ tr//"/;
        $line =~ tr//"/;
        $line =~ tr//-/;

        #======================================================================================
        # Ignore the prose parts
        if( !$inTokens )
        {
            if( $line =~ /11\.1[\s]General message structure/
                || $line =~ "USER_CREATED_FILE"
                || $line =~ "-- ASN1START" )
            
            {
                $inTokens = TRUE;
            }
            next;
        }
        else #inTokens is TRUE
        {
            if( $line =~ "-- ASN1STOP" )
            {
                $inTokens = FALSE;
                next;
            }
        }


        # These Word section header lines need to be skipped
        if( ($line =~ /11.2[\s]PDU definitions/)
            || ($line =~ /11.3[\s]Information element definitions/)
            || ($line =~ /11.4[\s]Constant definitions/)
            || ($line =~ /11.5[\s]RRC information between network nodes/) )
        {
            next;
        }
        #======================================================================================


        
        #======================================================================================
        # Begin parsing the good stuff.

        my @tokenarray = extractTokensFromLine($line);  # if blank line, 
                                                        # @tokenarray will contain a single FALSE

        # if the previous token is a BLOCK_COMMENT,
        # and if this line was a blank line or a BLOCK_COMMENT line,
        # then append this token's string to that one and continue to the next line
        if( ($#_tokens >= 0)
            && ($_tokens[$#_tokens]->{type} == ASN1Token::BLOCK_COMMENT)
            && (scalar(@tokenarray) == 1)
            && ( ($tokenarray[0]==FALSE) || ($tokenarray[0]->{type} == ASN1Token::BLOCK_COMMENT) )
        ) 
        {
            if( $tokenarray[0]==FALSE )
            {
                $_tokens[$#_tokens]->{string} .= "\n";
            }
            elsif( $tokenarray[0]->{type} == ASN1Token::BLOCK_COMMENT )
            {
                $_tokens[$#_tokens]->{string} .= $tokenarray[0]->{string};
            }

            next;
        }

        #skip blank lines (we are not in a block comment)
        if( (scalar(@tokenarray) == 1) && ($tokenarray[0]==FALSE) ) {
            next;
        }

        foreach(@tokenarray)
        {
            push @_tokens, $_;
        }
    }

    print " Done!  (",scalar(@_tokens)," tokens)\n\n";
}


# returns a list of (line, type, string) or FALSE
sub extractTokensFromLine($)
{
    my $eol_comment = "";

    $_ = $_[0];

    #if this line is blank, return FALSE and store no tokens
    if( /^[\s]$/ )
    {
        return FALSE;
    }

    #check if this line is a full-line comment; if yes, return it 
    if( /^[\s]*(--.*)/ )
    {
        return new ASN1Token( line=>$_inputline, type=>ASN1Token::BLOCK_COMMENT, string=>"$1\n" );
    }

    #look for a non-full line comment;
    # if yes, remove it and save for later (it has to be pushed last in the line)
    if( /(.*)(--.*)/ )
    {
        $_ = $1;
        $eol_comment = $2;
    }
    
    #remove leading/trailing whitespace
    /^[\s]*(.*)[\s]*$/;  
    $_ = $1;

    # split over whiteplace
    my @tokenstrings = split( /[\s]+/, $_ );

    my @returnarray;
    while(scalar(@tokenstrings))
    {
        $_ = shift @tokenstrings;

        #check if tokenstring contains non-space-separated tokens
        if( /(^.*?)([\(\)\{\},;]|\.\.\.|\.\.|::=)(.*$)/ )
        {
            # $1 = part before the special char (might be empty string)
            # $2 = the special char
            # $3 = the part after the special char (might be empty string)
            if( $1 ne "" )
            {
                push @returnarray, new ASN1Token( line=>$_inputline, type=>ASN1Token::TOKEN, string=>$1 );
            }
            push @returnarray, new ASN1Token( line=>$_inputline, type=>ASN1Token::TOKEN, string=>$2 );

            # more special characters might be present, so put it back for processing in the next loop iteration
            if( $3 ne "" )
            {
                unshift @tokenstrings, $3;
            }
        }
        else
        {
            push @returnarray, new ASN1Token( line=>$_inputline, type=>ASN1Token::TOKEN, string=>$_ );
        }
    }

    # if we found an eol_comment, push it now
    if($eol_comment)
    {
        push @returnarray, new ASN1Token( line=>$_inputline, type=>ASN1Token::EOL_COMMENT, string=>$eol_comment );
    }

    return @returnarray;
}


#------------------------------------
# returns the next token
#------------------------------------
sub getToken()
{
    if(@_tokens) {
        return shift @_tokens;
    }
    else {
        return FALSE;
    }
}

#------------------------------------
# Returns the xth next token
# but DOES NOT pop it from the stack
# (i.e., the xth call to getToken() will still give this token)
# Returns FALSE if less than x tokens remain on the stack.
#------------------------------------
sub peek(;$)
{
    my $x = 0;

    if( defined($_[0]) )
    {
        if( $_[0] <= 0 )
        {
            die("ASN1tokenizer::peek(): argument must be >0");
        }
        else
        {
            $x = $_[0]-1;
        }
    }

    if( $x <= $#_tokens )  #if x is <= than greatest index
    {
        return $_tokens[$x];
    }

    return FALSE;
}


1;
